;; Tin Can Communication - Transmitter
;; Email: mike@mikekohn.net
;;   Web: http://www.mikekohn.net/
;;
;; Receive data over UART and transmit as tones that will travel through
;; two tin cans connected with string and decode it with another circuit
;; that transmit the data over UART to the second computer.

.avr8
.include "tn2313def.inc"

; note: CLKSEL 0100  DIV 1
; octave:8> 1 / (1.1 * 5000 * 0.00000022)
; ans =  826.45 Hz
;
; 8,000,000 / 1,000 = 8000
; 8,000,000 / 826.45 = 9680
; 8,000,000 / 713.01 = 11220

;  8000 / 2 = 4000 (cycles per speaker toggle)
;  9680 / 2 = 4840 (cycles per speaker toggle)
; 11220 / 2 = 5610 (cycles per speaker toggle)

; Data is:
;   1 / 713 = 1.4ms
;   30ms START      | 240,000 cycles | 42 interrupts
;   20ms ONE        | 160,000 cycles | 28 interrupts
;   15ms ZERO       | 120,000 cycles | 21 interrupts
;   10ms SPACE      |  80,000 cycles | 14 interrupts

; Update 2020-June-28
; 10000 cycles between interrupts
; 0.00125 seconds between interrupts.
; Start = 42 interrupts = 0.052 seconds
;   One = 28 interrupts = 0.035 seconds
;  Zero = 21 interrupts = 0.02625 seconds
; Space = 14 interrupts = 0.0175 seconds

;BIT_START equ 42
;BIT_ONE   equ 28
;BIT_ZERO  equ 21
;BIT_SPACE equ 14

; Start = 160 interrupts = 0.200 seconds
;   One =  80 interrupts = 0.100 seconds
;  Zero =  60 interrupts = 0.075 seconds
; Space =  40 interrupts = 0.050 seconds

BIT_START equ 160
BIT_ONE   equ 80
BIT_ZERO  equ 60
BIT_SPACE equ 40 

;INTERRUPT_TIME equ 10000
INTERRUPT_TIME equ 9524

; r0  = 0
; r1  = 1
; r4  =
; r10 =
; r15 = 255
; r14 =
; r16 =
; r17 =
; r18 =
; r19 = 
; r20 = interrupt counter.
; r21 = Boolean: Should play tone.
; r22 = Current speaker value.
; r23 = UART data.
; r24 = transmit data
; r25 = bit counter
; r26 = data bit length
; r27 =
; r30 =
; r31 =

.org 0x0000
  rjmp start
.org 0x001a
  rjmp timer_interrupt

start:
  ;; Disable interrupts
  cli

  ldi r16, RAMEND & 0xff
  out SPL, r16

  ;; Setup registers
  eor r0, r0
  mov r1, r0
  inc r1
  mov r15, r0
  dec r15

  ;; Setup UART baud rate.
  out UBRRH, r0
  ldi r17, 51    ; 9600
  out UBRRL, r17 ; 51 @ 8MHz = 9600 baud

  ;; Setup UART options.
  ldi r17, (1 << UCSZ0) | (1 << UCSZ1) ; sets up data as 8N1
  out UCSRC, r17
  ldi r17, (1 << TXEN) | (1 << RXEN)   ; enables send/receive
  out UCSRB, r17
  out UCSRA, r0

  ;; Set up LED default
  ;; PB0 -> Debug LED.
  ;; PD5 -> Debug jumper (pulled down = debug beeping mode).
  ;; PD6 -> Sound out.
  ldi r17, (1 << PB0)
  out DDRB, r17
  out PORTB, r0
  ldi r17, (1 << PD6)
  out DDRD, r17
  ldi r17, (1 << PD6) | (1 << PD5)
  out PORTD, r17

  ; Set up TIMER1
  ldi r17, (INTERRUPT_TIME >> 8)
  out OCR1AH, r17
  ldi r17, (INTERRUPT_TIME & 0xff)
  out OCR1AL, r17
  ldi r17, (1 << OCIE1A)
  out TIMSK, r17        ; enable interrupt comare A 
  out TCCR1C, r0
  out TCCR1A, r0
  ldi r17, (1 << CS10) | (1 << WGM12)
  out TCCR1B, r17       ; No prescaler / OCR1A is top

  ; Enable interrupts
  sei

  ldi r21, 0

  sbis PIND, PD5
  rjmp blink_and_beep

main:
  ; Read character from UART
  sbis UCSRA, RXC
  rjmp main
  in r23, UDR

  mov r24, r23
  rcall send_data

  ; Echo char back for debug
main_echo_wait:
  sbis UCSRA, UDRE
  rjmp main_echo_wait
  out UDR, r23
  rjmp main

send_data:
  sbi PORTB, PB0
  ldi r21, 1
  ldi r20, 0
send_data_start_bit:
  cpi r20, BIT_START
  brne send_data_start_bit
  rcall send_space
  ldi r25, 8
send_data_next_bit:
  ldi r21, 1
  ldi r26, BIT_ONE
  sbrs r24, 7
  ldi r26, BIT_ZERO
  lsl r24
send_data_bit:
  cp r20, r26
  brne send_data_bit
  rcall send_space
  dec r25
  brne send_data_next_bit
  cbi PORTB, PB0
  ret

send_space:
  ldi r21, 0
  ldi r20, 0
send_space_delay:
  cpi r20, BIT_SPACE
  brne send_space_delay
  ret

blink_and_beep:
  ldi r21, 1
  sbi PORTB, PB0
  rcall delay_1s
  ldi r21, 0
  cbi PORTB, PB0
  rcall delay_1s
  rjmp blink_and_beep

delay_50ms:
  mov r20, r0
delay_50ms_loop:
  cpi r20, 83
  brne delay_50ms_loop
  ret

delay_1s:
  ldi r17, 20
delay_1s_loop:
  rcall delay_50ms
  dec r17
  brne delay_1s_loop
  ret

timer_interrupt:
  in r7, SREG
  inc r20
  cpi r21, 0
  brne timer_interrupt_play_tone
  cbi PORTD, PD6
  out SREG, r7
  reti
timer_interrupt_play_tone:
  sbrs r22, 0
  cbi PORTD, PD6
  sbrc r22, 0
  sbi PORTD, PD6
  inc r22
  out SREG, r7
  reti

